
"""
*****************************************************************
Licensed Materials - Property of IBM
(C) Copyright IBM Corp. 2019. All Rights Reserved.
US Government Users Restricted Rights - Use, duplication or
disclosure restricted by GSA ADP Schedule Contract with IBM Corp.
*****************************************************************
"""

import tensorflow as tf
import ddl
import numpy as np
import sys
from os import environ
import argparse
import time

print ("Executing eagerly? ", tf.executing_eagerly())
class Model(object):
    #num_grads is a list defining the shape of the gradients
    def __init__(self, num_grads, input_dtype):
        ones = tf.ones(shape=[1, num_grads], dtype=input_dtype)  * (5.0)
        self.W = tf.Variable(ones,trainable=True)

    def __call__(self, x):
        y = tf.zeros(x.shape, dtype=x.dtype)
        y = tf.reduce_sum(tf.matmul(x, self.W), axis=1, keepdims=True)
        return y


def loss (y_pred, y_act):
    return tf.reduce_mean(tf.square(y_pred - y_act))

def get_gradients(model, inputs, outputs):
    with tf.GradientTape() as t:
        current_loss = loss(model(inputs), outputs)
    return t.gradient(current_loss, model.W)

def train(model, inputs, outputs, epoch, learning_rate, average):
    dW = get_gradients(model, inputs, outputs)
    dW_n_v = ddl._grads_reduce([(dW, None)], average)


    dW, v = zip(*dW_n_v)
    dW = dW[0]

    model.W.assign_sub((learning_rate * dW))
    print("DDL={}, Epoch={}, dW={}, W={}".format
            (ddl.rank(), epoch, dW, model.W))
    return dW

def prepare_training_data (num_examples, input_dtype, rank):
    inputs = tf.ones(shape=[num_examples, 1], dtype=input_dtype)  * (rank + 1)
    outputs = np.linspace(0, rank+1, num=num_examples).reshape((num_examples, 1)) + inputs
    return (inputs, outputs)

def get_expected_grads(grad_size, num_grads, num_iters, learning_rate, input_dtype, average):
    result = np.zeros((num_iters, num_grads), dtype=input_dtype)
    model = Model(num_grads, input_dtype)

    for epoch in range(num_iters):
        grads = np.zeros((ddl.size(), num_grads), dtype=input_dtype)
        for rank in range(ddl.size()):
            (x, y) = prepare_training_data(grad_size, input_dtype, rank)

            grad = get_gradients(model, x, y)
            grads[rank] = grad

        result[epoch] += np.sum(grads, axis=0)
        if average:
            result[epoch] /= ddl.size()

        model.W.assign_sub((learning_rate * result[epoch].reshape(1, num_grads)))
    if ddl.rank() == 0:
        print ("expected grad={}".format(result))

    return result


def all_reduce_test (grad_size,
                     num_grads,
                     num_iters,
                     rank,
                     input_dtype,
                     average):
    print ("DDL {}: input_dtype={}, average={}, num_grads={}, grad_size={}".format(
                    ddl.rank(), input_dtype, average, num_grads, grad_size))

    #Create expected gradients for each iterations
    lr = 0.01
    expected_grads = get_expected_grads(grad_size, num_grads, num_iters, lr, input_dtype, average)

    #Get the model
    train_model = Model(num_grads, input_dtype)


    #Prepare training data
    (x, y) = prepare_training_data(grad_size, input_dtype, ddl.rank())

    actual_grads = np.zeros((num_iters, num_grads), dtype=input_dtype)
    for epoch in range(num_iters):
        actual_grads[epoch] = train(train_model, x, y, epoch, learning_rate=lr, average=average)

    if not np.allclose(actual_grads, expected_grads, rtol=1e-5):
        if ddl.rank() == 0:
            print ("DDL:{} printing (grad_reduce avg={}) grads={}".format(
                ddl.rank(),average,actual_grads))
            print ("DDL:{} ERROR: gradient from allreduce is not same as expected:{}".format
                    (ddl.rank(), expected_grads))
            sys.exit(1)
    print ("===================================================================")

if __name__ == '__main__':
    parser = argparse.ArgumentParser(
        description='DDL allreduce op testcase')
    parser.add_argument('--ddl_options', type=str,
                        default='-mode b:2x1 -dump_iter 100',
                        help='DDL options. e.g. "-mode b:2x1 -dump_iter 100"')
    parser.add_argument('--grad_size', type=int,
                        default=1024,
                        help='Size of each gradient')
    parser.add_argument('--num_grads', type=int,
                        default=4,
                        help='Number of gradients')
    parser.add_argument('--iterations', type=int,
                        default=5,
                        help='Number of iterations to perform')
    parser.add_argument('--average', type=bool,
                        default=True,
                        help='Use average for allreduce or not')
    parser.add_argument('--dtype', type=str,
                        default="float32",
                        help='Which data type to use for training')

    FLAGS, unparsed = parser.parse_known_args()

    if environ.get('DDL_OPTIONS') is None:
        #Need to explicitly call ddl.init
        ddl.init(FLAGS.ddl_options)

    ddl_rank    = ddl.rank() + 1
    grad_size   = FLAGS.grad_size
    num_grads   = FLAGS.num_grads
    num_iters   = FLAGS.iterations
    average     = FLAGS.average
    input_dtype = np.float16 if FLAGS.dtype == "float16" else np.float32

    all_reduce_test(grad_size,
                    num_grads=num_grads,
                    num_iters=num_iters,
                    rank=ddl_rank,
                    input_dtype=input_dtype,
                    average=average)
